#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# ==============================
# flowlib.py
# library for optical flow processing
# Author: Ruoteng Li
# Date: 6th Aug 2016
#Modified by Juan Montesinos  2018
# ==============================
"""
__author__ = "Juan Montesinos"
__copyright__ = "Copyright 2018"
__version__ = "0.1"
__maintainer__ = "Juan Montesinos"
__email__ = "juanfelipe.montesinos@upf.edu"
__status__ = "Stable"
import png
import numpy as np
import matplotlib.colors as cl
import matplotlib.pyplot as plt
from PIL import Image
import imageio
import os
import cv2
import math
import multiprocessing as mp
from skimage.transform import resize
from scipy import sparse

UNKNOWN_FLOW_THRESH = 1e7
SMALLFLOW = 0.0
LARGEFLOW = 1e8

"""
=============
Flow Section
=============
"""
__all__=['FlowResize','visualize_flow','read_flow','write_flow','fp2int','int2fp','colortest']
def _FlowResize3d(new_size,img,**kwargs):
    if img.ndim != 3:
        raise Exception('Wrong input size array')    
    H,W,C = img.shape
    if C !=2 and C !=3:
        raise Exception('Wrong input size array')
    
    img = resize(img,new_size,preserve_range=True,**kwargs)
    Rh = new_size[0]/H
    Rw = new_size[1]/W
    img[...,0] = img[...,0]*Rw
    img[...,1] = img[...,1]*Rh
    return img
#DEPRECATED
#def _FlowResize4d(new_size,img,**kwargs):
#    if img.ndim != 4:
#        raise Exception('Wrong input size array')
#    N,H,W,C = img.shape
#    new_size += (C,)
#    if C !=2 and C !=3:
#        raise Exception('Wrong input size array')
#    img = img.transpose(1,2,3,0) #Shape is now H,W,C,N
#    img = resize(img,new_size,preserve_range=True,**kwargs).transpose(3,0,1,2)
#    Rh = new_size[0]/H
#    Rw = new_size[1]/W
#    img[...,0] = img[...,0]*Rw
#    img[...,1] = img[...,1]*Rh
#    return img
def _FlowResize4d(new_size,img,**kwargs):
    if img.ndim != 4:
        raise Exception('Wrong input size array')
    resizer = FlowResize3d(new_size,**kwargs)
    output = np.stack(list(map(resizer,img)))       
    return output
def FlowResize(new_size,img,**kwargs):
    dims = img.ndim
    if dims == 4:
        return _FlowResize4d(new_size,img,**kwargs)
    elif dims == 3:
        return _FlowResize3d(new_size,img,**kwargs)
    else:
        raise Exception('Wrong input size array')
        
class FlowResize3d(object):
    def __init__(self,new_size,**kwargs):
        self.new_size = new_size
        self.kwargs = kwargs
    def __call__(self,img):
        return FlowResize(self.new_size,img,**self.kwargs)
    def __repr__(self):
        return 'Optical Flow Resize: {0}'.format(self.new_size)        
class _FlowResize(object):
    def __init__(self,new_size,**kwargs):
        self.new_size = new_size
        self.kwargs = kwargs
    def __call__(self,img):
        return FlowResize(self.new_size,img,**self.kwargs)
    def __repr__(self):
        return 'Optical Flow Resize: {0}'.format(self.new_size)
def show_flow(filename):
    """
    visualize optical flow map using matplotlib
    :param filename: optical flow file
    :return: None
    """
    flow = read_flow(filename)   
    img = flow_to_image(flow[...,:2])
    plt.imshow(img)
    plt.show()


def visualize_flow(flow, mode='Y'):
    """
    this function visualize the input flow
    :param flow: input flow in array
    :param mode: choose which color mode to visualize the flow (Y: Ccbcr, RGB: RGB color)
    :return: None
    """
    if mode == 'Y':
        # Ccbcr color wheel
        img = flow_to_image(flow)
        plt.imshow(img)
        plt.show()
    elif mode == 'RGB':
        (h, w) = flow.shape[0:2]
        du = flow[:, :, 0]
        dv = flow[:, :, 1]
        valid = flow[:, :, 2]
        max_flow = max(np.max(du), np.max(dv))
        img = np.zeros((h, w, 3), dtype=np.float64)
        # angle layer
        img[:, :, 0] = np.arctan2(dv, du) / (2 * np.pi)
        # magnitude layer, normalized to 1
        img[:, :, 1] = np.sqrt(du * du + dv * dv) * 8 / max_flow
        # phase layer
        img[:, :, 2] = 8 - img[:, :, 1]
        # clip to [0,1]
        small_idx = img[:, :, 0:3] < 0
        large_idx = img[:, :, 0:3] > 1
        img[small_idx] = 0
        img[large_idx] = 1
        # convert to rgb
        img = cl.hsv_to_rgb(img)
        # remove invalid point
        img[:, :, 0] = img[:, :, 0] * valid
        img[:, :, 1] = img[:, :, 1] * valid
        img[:, :, 2] = img[:, :, 2] * valid
        # show
        plt.imshow(img)
        plt.show()

    return None
def fp2int(x,ui,precision): #Recomended precision: 64, Allows to encode flows between [-2,2] with an accuracy 1/64
    uis= str(ui)
    return np.maximum(np.minimum(x*precision+2.0**(ui-1),2.0**ui-1),0).astype(np.dtype('uint'+uis))
def int2fp(x,ui,precision):
    return ((x - 2.0 ** (ui-1)) / precision)
def read_flow_flo(filename):
    """
    Modified by Juan Montesinos
    read optical flow from Middlebury .flo file
    :param filename: name of the flow file
    :return: optical flow data in matrix
    """
    f = open(filename, 'rb')
    try:
        magic = np.fromfile(f, np.float32, count=1)[0]    # For Python3.x
    except:
        magic = np.fromfile(f, np.float32, count=1)       # For Python2.x
    data2d = None

    if 202021.25 != magic:
        print('Magic number incorrect. Invalid .flo file')
    else:
        try:
            w = np.fromfile(f, np.int32, count=1)[0]     # For Python3.x
            h = np.fromfile(f, np.int32, count=1)[0]     # For Python2.x
        except:
            w = np.fromfile(f, np.int32, count=1)
            h = np.fromfile(f, np.int32, count=1)
        print("Reading %d x %d flo file" % (h, w))
        data2d = np.fromfile(f, np.float32, count=2 * w * h)
        # reshape data into 3D array (columns, rows, channels)
        data2d = np.resize(data2d, (h, w, 2))
    f.close()
    return data2d




def read_flow_png(flow_file,precision=64):
    """
    Read optical flow from KITTI .png file
    :param flow_file: name of the flow file
    :return: optical flow data in matrix
    """
    flow_object = png.Reader(filename=flow_file)
    flow_direct = flow_object.asDirect()
    flow_data = list(flow_direct[2])
    (w, h) = flow_direct[3]['size']
    planes = flow_direct[3]['planes']
    flow = np.zeros((h, w, planes), dtype=np.float64)
    custom = planes==2
    for i in range(len(flow_data)):
        flow[i, :, 0] = flow_data[i][0::planes]
        flow[i, :, 1] = flow_data[i][1::planes]
        if not custom:
            flow[i, :, 2] = flow_data[i][2::planes]

    flow[:, :, 0:2] = (flow[:, :, 0:2] - 2 ** 15) / precision
    if not custom:
        invalid_idx = (flow[:, :, 2] == 0)
        flow[invalid_idx, 0] = 0
        flow[invalid_idx, 1] = 0
    return flow
def read_flow_npy_sparse(flow_file):
    flow = np.load(flow_file).item()
    flow = np.array(flow.todense())
    s = flow.shape
    flow = flow.reshape(s[0],int(s[1]/2),2)
    return flow
def read_flow_core(flow_file,*args,**kwargs):
    filename,extension = os.path.splitext(flow_file)
    if extension =='.png':
        return read_flow_png(flow_file,*args,**kwargs)
    elif extension == '.jpg' or extension == '.jpeg':
        x = imageio.imread(flow_file)
        return int2fp(x,ui=8,*args,**kwargs)
    elif extension == '.flo':
        return read_flow_flo(flow_file)
    elif extension == '.npy':
        return read_flow_npy_sparse(flow_file)
    else:
        raise Exception('Non recognized file format, use .jpg, .jpeg, .flo, .png')

def read_flow(flow_files,multiprocessing=0,*args,**kwargs):
    if isinstance(flow_files,str) and os.path.isfile(flow_files):
        return read_flow_core(flow_files,*args,**kwargs)
    elif (isinstance(flow_files,str) and not os.path.isfile(flow_files)) or isinstance(flow_files,list):
        if isinstance(flow_files,str):
            files = sorted(os.listdir(flow_files))
            flow_files = [os.path.join(flow_files,file) for file in files]           
        if multiprocessing != 0:
            pool = mp.Pool(multiprocessing)
            results = [pool.apply(read_flow_core, args =(file,*args),kwargs=kwargs) for file in flow_files]
            pool.close()
        else:
            results = [read_flow_core(file,*args,**kwargs) for file in flow_files]
        return np.stack(results)
    else:
        raise Exception('Read_flow requires file_path or list of ordered file_paths as input')
def write_flow_png(flow,filename,precision,error=False):
    precision=float(precision)
    """
    Added by Juan Montesinos
    A random example shows a mean error of 0.0078 with std 0.0045
    write optical flow in 16bit 2-channel png file (grayscale + alpha)
    :param flow: optical flow map
    :param filename: optical flow file path to be saved
    :return: None
    """
    if not (flow.dtype == 'float32' or flow.dtype == 'float64'):
        raise ValueError('Float data required')
    if len(flow.shape) != 3:
        raise Exception('Input has more than 2 dimensions, wrong flow format')
    if flow.shape[2] != 2:
        raise Exception('Unexpected amount of channels: {0}. Expecting two channels u,v').format(flow.shape[2])
    flow = flow.astype(np.float64)

    (H,W,C) = flow.shape
    out = np.zeros((H,W,2),dtype=np.uint16)
    out[:,:,0] =  np.maximum(np.minimum(flow[:,:,0]*precision+2.0**15,2.0**16-1),0).astype(np.uint16)
    out[:,:,1] =  np.maximum(np.minimum(flow[:,:,1]*precision+2.0**15,2.0**16-1),0).astype(np.uint16)
    png.from_array(out,mode='LA').save(os.path.splitext(filename)[0]+'.png')
    if error:
        flow= out.astype(np.float32)   
        flow[:, :, 0:2] = (flow[:, :, 0:2] - 2 ** 15) / precision            
        return flow
def write_flow_flo(flow, filename):
    """
    write optical flow in Middlebury .flo format
    :param flow: optical flow map
    :param filename: optical flow file path to be saved
    :return: None
    """
    flow = flow[...,:2]
    f = open(filename, 'wb')
    magic = np.array([202021.25], dtype=np.float32)
    (height, width) = flow.shape[0:2]
    w = np.array([width], dtype=np.int32)
    h = np.array([height], dtype=np.int32)
    magic.tofile(f)
    w.tofile(f)
    h.tofile(f)
    flow.tofile(f)
    f.close()
    
def write_flow_npy_sparse(flow,filename,th):
    """
    Write optical flow as numpy files containing sparse matrices
    :param flow: optical flow map
    :param filename: optical flow file path to be saved
    :param th: In order to convert optical flows in sparse matrices, pixels 
    whose magnitude < th will be zeroed.
    """
    flow = flow*(np.linalg.norm(flow,axis=2)>th)[...,None]
    flow=flow.reshape(flow.shape[0],-1)
    flow = sparse.csr_matrix(flow)
    np.save(filename,flow)
def write_flow(flow,filename,**kwargs):
    """
    Write optical flow in png, flo or jpg format.
    Note:
    PNG saving process is ~6 times slower. It maps optical flow to
    uin16 and saves 2-channel png (gray + alpha).
    
    JPG maps flow to uint8. Uses faster scipy libraries. A  3rd channel is free
    to be used, else it will be filled with zeros.
    JPG is a compressive method which requires ~25 times less memory.
    
    FLO is the standard flo format for optical flow.
    """
    root,extension = os.path.splitext(filename)
    if extension =='.png':
        return write_flow_png(flow,filename,kwargs['precision'])
    elif extension == '.jpg' or extension == '.jpeg':
        
        flow = fp2int(flow,8,kwargs['precision'])
        (H,W,C) = flow.shape
        if C == 2:
            zeros = np.zeros((H,W,1),dtype=np.uint8)
            flow = np.concatenate([flow,zeros],axis=2)

            
        imageio.imwrite(filename,flow)        
    elif extension == '.flo':
        write_flow_flo(flow,filename,*kwargs)
    elif extension == '.npy':
        write_flow_npy_sparse(flow,filename,kwargs['th'])
    else:
        raise Exception('Non recognized file format, use .jpg, .jpeg, .flo, .png, .npy')    
  
def segment_flow(flow):
    h = flow.shape[0]
    w = flow.shape[1]
    u = flow[:, :, 0]
    v = flow[:, :, 1]

    idx = ((abs(u) > LARGEFLOW) | (abs(v) > LARGEFLOW))
    idx2 = (abs(u) == SMALLFLOW)
    class0 = (v == 0) & (u == 0)
    u[idx2] = 0.00001
    tan_value = v / u

    class1 = (tan_value < 1) & (tan_value >= 0) & (u > 0) & (v >= 0)
    class2 = (tan_value >= 1) & (u >= 0) & (v >= 0)
    class3 = (tan_value < -1) & (u <= 0) & (v >= 0)
    class4 = (tan_value < 0) & (tan_value >= -1) & (u < 0) & (v >= 0)
    class8 = (tan_value >= -1) & (tan_value < 0) & (u > 0) & (v <= 0)
    class7 = (tan_value < -1) & (u >= 0) & (v <= 0)
    class6 = (tan_value >= 1) & (u <= 0) & (v <= 0)
    class5 = (tan_value >= 0) & (tan_value < 1) & (u < 0) & (v <= 0)

    seg = np.zeros((h, w))

    seg[class1] = 1
    seg[class2] = 2
    seg[class3] = 3
    seg[class4] = 4
    seg[class5] = 5
    seg[class6] = 6
    seg[class7] = 7
    seg[class8] = 8
    seg[class0] = 0
    seg[idx] = 0

    return seg


def flow_error(tu, tv, u, v):
    """
    Calculate average end point error
    :param tu: ground-truth horizontal flow map
    :param tv: ground-truth vertical flow map
    :param u:  estimated horizontal flow map
    :param v:  estimated vertical flow map
    :return: End point error of the estimated flow
    """
    smallflow = 0.0
    '''
    stu = tu[bord+1:end-bord,bord+1:end-bord]
    stv = tv[bord+1:end-bord,bord+1:end-bord]
    su = u[bord+1:end-bord,bord+1:end-bord]
    sv = v[bord+1:end-bord,bord+1:end-bord]
    '''
    stu = tu[:]
    stv = tv[:]
    su = u[:]
    sv = v[:]

    idxUnknow = (abs(stu) > UNKNOWN_FLOW_THRESH) | (abs(stv) > UNKNOWN_FLOW_THRESH)
    stu[idxUnknow] = 0
    stv[idxUnknow] = 0
    su[idxUnknow] = 0
    sv[idxUnknow] = 0

    ind2 = [(np.absolute(stu) > smallflow) | (np.absolute(stv) > smallflow)]
    index_su = su[ind2]
    index_sv = sv[ind2]
    an = 1.0 / np.sqrt(index_su ** 2 + index_sv ** 2 + 1)
    un = index_su * an
    vn = index_sv * an

    index_stu = stu[ind2]
    index_stv = stv[ind2]
    tn = 1.0 / np.sqrt(index_stu ** 2 + index_stv ** 2 + 1)
    tun = index_stu * tn
    tvn = index_stv * tn

    '''
    angle = un * tun + vn * tvn + (an * tn)
    index = [angle == 1.0]
    angle[index] = 0.999
    ang = np.arccos(angle)
    mang = np.mean(ang)
    mang = mang * 180 / np.pi
    '''

    epe = np.sqrt((stu - su) ** 2 + (stv - sv) ** 2)
    epe = epe[ind2]
    mepe = np.mean(epe)
    return mepe


def flow_to_image(flow, display=False):
    """
    Convert flow into middlebury color code image
    :param flow: optical flow map
    :return: optical flow image in middlebury color
    """
    u = flow[:, :, 0]
    v = flow[:, :, 1]

    maxu = -999.
    maxv = -999.
    minu = 999.
    minv = 999.

    idxUnknow = (abs(u) > UNKNOWN_FLOW_THRESH) | (abs(v) > UNKNOWN_FLOW_THRESH)
    u[idxUnknow] = 0
    v[idxUnknow] = 0

    maxu = max(maxu, np.max(u))
    minu = min(minu, np.min(u))

    maxv = max(maxv, np.max(v))
    minv = min(minv, np.min(v))

    rad = np.sqrt(u ** 2 + v ** 2)
    maxrad = max(-1, np.max(rad))

    if display:
        print("max flow: %.4f\nflow range:\nu = %.3f .. %.3f\nv = %.3f .. %.3f" % (maxrad, minu,maxu, minv, maxv))

    u = u/(maxrad + np.finfo(float).eps)
    v = v/(maxrad + np.finfo(float).eps)

    img = compute_color(u, v)

    idx = np.repeat(idxUnknow[:, :, np.newaxis], 3, axis=2)
    img[idx] = 0

    return np.uint8(img)


def evaluate_flow_file(gt, pred):
    """
    evaluate the estimated optical flow end point error according to ground truth provided
    :param gt: ground truth file path
    :param pred: estimated optical flow file path
    :return: end point error, float32
    """
    # Read flow files and calculate the errors
    gt_flow = read_flow(gt)        # ground truth flow
    eva_flow = read_flow(pred)     # predicted flow
    # Calculate errors
    average_pe = flow_error(gt_flow[:, :, 0], gt_flow[:, :, 1], eva_flow[:, :, 0], eva_flow[:, :, 1])
    return average_pe


def evaluate_flow(gt_flow, pred_flow):
    """
    gt: ground-truth flow
    pred: estimated flow
    """
    average_pe = flow_error(gt_flow[:, :, 0], gt_flow[:, :, 1], pred_flow[:, :, 0], pred_flow[:, :, 1])
    return average_pe


"""
==============
Disparity Section
==============
"""


def read_disp_png(file_name):
    """
    Read optical flow from KITTI .png file
    :param file_name: name of the flow file
    :return: optical flow data in matrix
    """
    image_object = png.Reader(filename=file_name)
    image_direct = image_object.asDirect()
    image_data = list(image_direct[2])
    (w, h) = image_direct[3]['size']
    channel = len(image_data[0]) / w
    flow = np.zeros((h, w, channel), dtype=np.uint16)
    for i in range(len(image_data)):
        for j in range(channel):
            flow[i, :, j] = image_data[i][j::channel]
    return flow[:, :, 0] / 256


def disp_to_flowfile(disp, filename):
    """
    Read KITTI disparity file in png format
    :param disp: disparity matrix
    :param filename: the flow file name to save
    :return: None
    """
    f = open(filename, 'wb')
    magic = np.array([202021.25], dtype=np.float32)
    (height, width) = disp.shape[0:2]
    w = np.array([width], dtype=np.int32)
    h = np.array([height], dtype=np.int32)
    empty_map = np.zeros((height, width), dtype=np.float32)
    data = np.dstack((disp, empty_map))
    magic.tofile(f)
    w.tofile(f)
    h.tofile(f)
    data.tofile(f)
    f.close()


"""
==============
Image Section
==============
"""


def read_image(filename):
    """
    Read normal image of any format
    :param filename: name of the image file
    :return: image data in matrix uint8 type
    """
    img = Image.open(filename)
    im = np.array(img)
    return im


def warp_image(im, flow):
    """
    Use optical flow to warp image to the next
    :param im: image to warp
    :param flow: optical flow
    :return: warped image
    """
    from scipy import interpolate
    image_height = im.shape[0]
    image_width = im.shape[1]
    flow_height = flow.shape[0]
    flow_width = flow.shape[1]
    n = image_height * image_width
    (iy, ix) = np.mgrid[0:image_height, 0:image_width]
    (fy, fx) = np.mgrid[0:flow_height, 0:flow_width]
    fx += flow[:,:,0]
    fy += flow[:,:,1]
    mask = np.logical_or(fx <0 , fx > flow_width)
    mask = np.logical_or(mask, fy < 0)
    mask = np.logical_or(mask, fy > flow_height)
    fx = np.minimum(np.maximum(fx, 0), flow_width)
    fy = np.minimum(np.maximum(fy, 0), flow_height)
    points = np.concatenate((ix.reshape(n,1), iy.reshape(n,1)), axis=1)
    xi = np.concatenate((fx.reshape(n, 1), fy.reshape(n,1)), axis=1)
    warp = np.zeros((image_height, image_width, im.shape[2]))
    for i in range(im.shape[2]):
        channel = im[:, :, i]
        plt.imshow(channel, cmap='gray')
        values = channel.reshape(n, 1)
        new_channel = interpolate.griddata(points, values, xi, method='cubic')
        new_channel = np.reshape(new_channel, [flow_height, flow_width])
        new_channel[mask] = 1
        warp[:, :, i] = new_channel.astype(np.uint8)

    return warp.astype(np.uint8)


"""
==============
Others
==============
"""

def scale_image(image, new_range):
    """
    Linearly scale the image into desired range
    :param image: input image
    :param new_range: the new range to be aligned
    :return: image normalized in new range
    """
    min_val = np.min(image).astype(np.float32)
    max_val = np.max(image).astype(np.float32)
    min_val_new = np.array(min(new_range), dtype=np.float32)
    max_val_new = np.array(max(new_range), dtype=np.float32)
    scaled_image = (image - min_val) / (max_val - min_val) * (max_val_new - min_val_new) + min_val_new
    return scaled_image


def compute_color(u, v):
    """
    compute optical flow color map
    :param u: optical flow horizontal map
    :param v: optical flow vertical map
    :return: optical flow in color code
    """
    [h, w] = u.shape
    img = np.zeros([h, w, 3])
    nanIdx = np.isnan(u) | np.isnan(v)
    u[nanIdx] = 0
    v[nanIdx] = 0

    colorwheel = make_color_wheel()
    ncols = np.size(colorwheel, 0)

    rad = np.sqrt(u**2+v**2)

    a = np.arctan2(-v, -u) / np.pi

    fk = (a+1) / 2 * (ncols - 1) + 1

    k0 = np.floor(fk).astype(int)

    k1 = k0 + 1
    k1[k1 == ncols+1] = 1
    f = fk - k0

    for i in range(0, np.size(colorwheel,1)):
        tmp = colorwheel[:, i]
        col0 = tmp[k0-1] / 255
        col1 = tmp[k1-1] / 255
        col = (1-f) * col0 + f * col1

        idx = rad <= 1
        col[idx] = 1-rad[idx]*(1-col[idx])
        notidx = np.logical_not(idx)

        col[notidx] *= 0.75
        img[:, :, i] = np.uint8(np.floor(255 * col*(1-nanIdx)))

    return img
def colortest():
    truerange = 1
    height = 151
    width  = 151
    range_f = truerange * 1.04
    
    s2 = int(round(height/2))
    x, y = np.meshgrid(np.arange(1, height + 1, 1), np.arange(1, width + 1, 1))
    
    u = x*range_f/s2 - range_f
    v = y*range_f/s2 - range_f
    
    img = computeColor(u/truerange, v/truerange)
    
    img[s2,:,:] = 0
    img[:,s2,:] = 0
    
    cv2.imshow('test color pattern',img)
    k = cv2.waitKey()
    
    F = np.stack((u, v), axis = 2)
    write_flow(F, 'colorTest.flo')
    
    flow = read_flow('colorTest.flo')
    
    u = flow[: , : , 0]
    v = flow[: , : , 1]		
    
    img = computeColor.computeColor(u/truerange, v/truerange)
    
    img[s2,:,:] = 0
    img[:,s2,:] = 0
    
    cv2.imshow('saved and reloaded test color pattern',img)
    k = cv2.waitKey()
    
    # color encoding scheme for optical flow
    img = computeColor.computeColor(u/range_f/math.sqrt(2), v/range_f/math.sqrt(2));
    
    cv2.imshow('optical flow color encoding scheme',img)
    cv2.imwrite('colorTest.png', img)
    k = cv2.waitKey()
def computeColor(u, v):

	colorwheel = make_color_wheel();
	nan_u = np.isnan(u)
	nan_v = np.isnan(v)
	nan_u = np.where(nan_u)
	nan_v = np.where(nan_v) 

	u[nan_u] = 0
	u[nan_v] = 0
	v[nan_u] = 0 
	v[nan_v] = 0

	ncols = colorwheel.shape[0]
	radius = np.sqrt(u**2 + v**2)
	a = np.arctan2(-v, -u) / np.pi
	fk = (a+1) /2 * (ncols-1) # -1~1 maped to 1~ncols
	k0 = fk.astype(np.uint8)	 # 1, 2, ..., ncols
	k1 = k0+1;
	k1[k1 == ncols] = 0
	f = fk - k0

	img = np.empty([k1.shape[0], k1.shape[1],3])
	ncolors = colorwheel.shape[1]
	for i in range(ncolors):
		tmp = colorwheel[:,i]
		col0 = tmp[k0]/255
		col1 = tmp[k1]/255
		col = (1-f)*col0 + f*col1
		idx = radius <= 1
		col[idx] = 1 - radius[idx]*(1-col[idx]) # increase saturation with radius    
		col[~idx] *= 0.75 # out of range
		img[:,:,2-i] = np.floor(255*col).astype(np.uint8)

	return img.astype(np.uint8)
def make_color_wheel():
    """
    Generate color wheel according Middlebury color code
    :return: Color wheel
    """
    RY = 15
    YG = 6
    GC = 4
    CB = 11
    BM = 13
    MR = 6

    ncols = RY + YG + GC + CB + BM + MR

    colorwheel = np.zeros([ncols, 3])

    col = 0

    # RY
    colorwheel[0:RY, 0] = 255
    colorwheel[0:RY, 1] = np.transpose(np.floor(255*np.arange(0, RY) / RY))
    col += RY

    # YG
    colorwheel[col:col+YG, 0] = 255 - np.transpose(np.floor(255*np.arange(0, YG) / YG))
    colorwheel[col:col+YG, 1] = 255
    col += YG

    # GC
    colorwheel[col:col+GC, 1] = 255
    colorwheel[col:col+GC, 2] = np.transpose(np.floor(255*np.arange(0, GC) / GC))
    col += GC

    # CB
    colorwheel[col:col+CB, 1] = 255 - np.transpose(np.floor(255*np.arange(0, CB) / CB))
    colorwheel[col:col+CB, 2] = 255
    col += CB

    # BM
    colorwheel[col:col+BM, 2] = 255
    colorwheel[col:col+BM, 0] = np.transpose(np.floor(255*np.arange(0, BM) / BM))
    col += + BM

    # MR
    colorwheel[col:col+MR, 2] = 255 - np.transpose(np.floor(255 * np.arange(0, MR) / MR))
    colorwheel[col:col+MR, 0] = 255

    return colorwheel

